import json
import logging
import re

import pandas
from slugify import slugify

from nibe.heatpump import Model

logger = logging.getLogger("nibe").getChild(__name__)


class CSVConverter:
    def __init__(self, in_file, out_file):
        self.in_file = in_file
        self.out_file = out_file

        self.data = None

    def convert(self):
        self._read_csv()

        self._lowercase_column_names()

        self._update_index()

        self._make_name_using_slugify()

        self._fix_data_unit_column()

        self._fix_data_types()

        self._replace_mode_with_boolean_write_parameter()

        self._unset_equal_min_max_default_values()

        self._make_mapping_parameter()

        self._export_to_file()

    def _make_dict(self):
        return {index: row.dropna().to_dict() for index, row in self.data.iterrows()}

    def _make_mapping_parameter(self):
        re_mapping = re.compile(
            r"(?P<value>\d+|I)\s*=\s*(?P<key>(?:[\w +.-]+[\w]\b[+]?(?! *=)))",
            re.IGNORECASE,
        )
        mappings = (
            self.data["info"]
            .where(~self.data["info"].str.contains("encoded"))
            .str.extractall(re_mapping)
        )
        mappings["value"] = mappings["value"].str.replace("I", "1").astype("int")
        mappings = mappings.reset_index("match", drop=True)
        mappings = mappings.drop_duplicates()
        self.data["mappings"] = pandas.Series(
            {k: dict(g.values) for k, g in mappings.groupby("value", level=0)}
        )

    def _unset_equal_min_max_default_values(self):
        valid_min_max = self.data["min"] != self.data["max"]
        self.data["min"] = self.data["min"].where(valid_min_max)
        self.data["max"] = self.data["max"].where(valid_min_max)
        self.data["default"] = self.data["default"].where(valid_min_max)

    def _fix_data_types(self):
        self.data["unit"] = self.data["unit"].astype("string")
        self.data["title"] = self.data["title"].astype("string")
        self.data["info"] = self.data["info"].astype("string")
        self.data["size"] = self.data["size"].astype("string")
        self.data["name"] = self.data["name"].astype("string")

    def _fix_data_unit_column(self):
        self.data["unit"] = (
            self.data["unit"].replace(r"^\s*$", pandas.NA, regex=True).str.strip()
        )

    def _make_name_using_slugify(self):
        ids = pandas.Series(self.data.index, index=self.data.index)
        self.data["name"] = self.data["title"].combine(
            ids, lambda title, id_: slugify(f"{title}-{id_}")
        )

    def _replace_mode_with_boolean_write_parameter(self):
        self.data["mode"] = self.data["mode"].str.strip().astype("string")

        self.data["write"] = self.data["mode"].map(
            lambda x: True if x == "R/W" else pandas.NA
        )
        del self.data["mode"]

    def _lowercase_column_names(self):
        self.data.columns = map(str.lower, self.data.columns)

    def _read_csv(self):
        self.data = pandas.read_csv(
            self.in_file,
            sep=";",
            skiprows=4,
            encoding="latin1",
            index_col=False,
            skipinitialspace=True,
        )

    def _update_index(self):
        self.data = self.data.set_index("id")

    def _export_to_file(self):
        o = self._make_dict()
        with open(self.out_file, "w") as fh:
            json.dump(o, fh, indent=2)


def run():
    for in_file in Model.get_data_path().glob("*.csv"):
        out_file = in_file.with_suffix(".json")
        logger.info(f"Converting {in_file} to {out_file}")
        CSVConverter(in_file, out_file).convert()
        logger.info(f"Converted {in_file} to {out_file}")


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    run()



import asyncio
import logging
import socket
from asyncio import CancelledError, InvalidStateError
from binascii import hexlify
from contextlib import suppress
from functools import reduce
from io import BytesIO
from operator import xor

from construct import (Array, Bytes, Checksum, ChecksumError, Const, Enum, FixedSized,
                       Flag, Int8ub, Int16ul, RawCopy, Struct, Subconstruct, Switch,
                       this,)

from nibe.coil import Coil
from nibe.connection import Connection
from nibe.exceptions import (CoilNotFoundException, CoilReadException,
                             CoilReadTimeoutException, CoilWriteException,
                             CoilWriteTimeoutException, DecodeException, NibeException,)
from nibe.heatpump import HeatPump

logger = logging.getLogger("nibe").getChild(__name__)


class NibeGW(asyncio.DatagramProtocol, Connection):
    DEFAULT_TIMEOUT = 5

    def __init__(
        self,
        heatpump: HeatPump,
        remote_ip: str,
        remote_read_port: int = 9999,
        remote_write_port: int = 10000,
        listening_ip: str = "0.0.0.0",
        listening_port: int = 9999,
    ) -> None:
        self._heatpump = heatpump
        self._listening_ip = listening_ip
        self._listening_port = listening_port

        self._remote_ip = remote_ip
        self._remote_read_port = remote_read_port
        self._remote_write_port = remote_write_port

        self._transport = None

        self._send_lock = asyncio.Lock()
        self._write_future = None
        self._read_future = None

    async def start(self):
        logger.info(f"Starting UDP server on port {self._listening_port}")

        await asyncio.get_event_loop().create_datagram_endpoint(
            lambda: self,
            local_addr=(self._listening_ip, self._listening_port),
            proto=socket.IPPROTO_UDP,
        )

    def connection_made(self, transport):
        self._transport = transport

    def datagram_received(self, data, addr):
        logger.debug(f"Received {hexlify(data)} from {addr}")
        try:
            msg = Response.parse(data)
            logger.debug(msg)
            cmd = msg.fields.value.cmd
            if cmd == "MODBUS_DATA_MSG":
                for row in msg.fields.value.data:
                    try:
                        self._on_raw_coil_value(row.coil_address, row.value)
                    except NibeException as e:
                        logger.error(str(e))
            elif cmd == "MODBUS_READ_RESP":
                row = msg.fields.value.data
                try:
                    self._on_raw_coil_value(row.coil_address, row.value)
                    with suppress(InvalidStateError, CancelledError, AttributeError):
                        self._read_future.set_result(None)
                except NibeException as e:
                    with suppress(InvalidStateError, CancelledError, AttributeError):
                        self._read_future.set_exception(CoilReadException(str(e), e))
                    raise
            elif cmd == "MODBUS_WRITE_RESP":
                with suppress(InvalidStateError, CancelledError, AttributeError):
                    self._write_future.set_result(msg.fields.value.data.result)
            else:
                logger.debug(f"Unknown command {cmd}")
        except ChecksumError:
            logger.warning(
                f"Ignoring packet from {addr} due to checksum error: {hexlify(data)}"
            )
        except NibeException as e:
            logger.error(f"Failed handling packet from {addr}: {e}")
        except Exception as e:
            logger.exception(
                f"Unexpected exception during parsing packet data '{hexlify(data)}' from {addr}",
                e,
            )

    async def read_coil(self, coil: Coil, timeout: float = DEFAULT_TIMEOUT) -> Coil:
        async with self._send_lock:
            data = ReadRequest.build(
                dict(fields=dict(value=dict(coil_address=coil.address)))
            )

            self._read_future = asyncio.get_event_loop().create_future()

            logger.debug(
                f"Sending {hexlify(data)} (read request) to {self._remote_ip}:{self._remote_write_port}"
            )
            self._transport.sendto(data, (self._remote_ip, self._remote_read_port))
            logger.debug(f"Waiting for read response for {coil.name}")

            try:
                await asyncio.wait_for(self._read_future, timeout)
            except asyncio.TimeoutError:
                raise CoilReadTimeoutException(
                    f"Timeout waiting for read response for {coil.name}"
                )
            finally:
                self._read_future = None

            return coil

    async def write_coil(self, coil: Coil, timeout: float = DEFAULT_TIMEOUT) -> Coil:
        assert coil.is_writable, f"{coil.name} is not writable"
        assert coil.value is not None
        async with self._send_lock:
            data = WriteRequest.build(
                dict(
                    fields=dict(
                        value=dict(coil_address=coil.address, value=coil.raw_value)
                    )
                )
            )

            self._write_future = asyncio.get_event_loop().create_future()

            logger.debug(
                f"Sending {hexlify(data)} (write request) to {self._remote_ip}:{self._remote_write_port}"
            )
            self._transport.sendto(data, (self._remote_ip, self._remote_write_port))

            try:
                await asyncio.wait_for(self._write_future, timeout)

                result = self._write_future.result()

                if not result:
                    raise CoilWriteException(f"Heatpump denied writing {coil.name}")
                else:
                    logger.info(f"Write succeeded for {coil.name}")
            except asyncio.TimeoutError:
                raise CoilWriteTimeoutException(
                    f"Timeout waiting for write feedback for {coil.name}"
                )
            finally:
                self._write_future = None

            return coil

    def error_received(self, exc):
        logger.error(exc)

    def _on_raw_coil_value(self, coil_address: int, raw_value: bytes):
        try:
            coil = self._heatpump.get_coil_by_address(coil_address)
        except CoilNotFoundException:
            if coil_address == 65535:  # 0xffff
                return
            raise

        coil.raw_value = raw_value
        logger.info(f"{coil.name}: {coil.value}")
        self._heatpump.notify_coil_update(coil)

    async def stop(self):
        self._transport.close()
        self._transport = None


def xor8(data: bytes) -> int:
    chksum = reduce(xor, data)
    if chksum == 0x5C:
        chksum = 0xC5
    return chksum


class Dedupe5C(Subconstruct):
    def __init__(self, subcon):
        super().__init__(subcon)
        self.name = subcon.name

    def _parse(self, stream, context, path):
        unescaped = stream.getvalue().replace(b"\x5c\x5c", b"\x5c")
        context.length = len(unescaped)
        with BytesIO(unescaped) as stream2:
            obj = self.subcon._parsereport(stream2, context, path)
        return obj

    def _build(self, obj, stream, context, path):
        escaped = stream.getvalue().replace(b"\x5c", b"\x5c\x5c")
        context.length = len(escaped)
        with BytesIO(escaped) as stream2:
            buildret = self.subcon._build(obj, stream2, context, path)
        return obj


Data = Dedupe5C(
    Switch(
        this.cmd,
        {
            "MODBUS_READ_RESP": Struct("coil_address" / Int16ul, "value" / Bytes(4)),
            "MODBUS_DATA_MSG": Array(
                lambda this: this.length // 4,
                Struct("coil_address" / Int16ul, "value" / Bytes(2)),
            ),
            "MODBUS_WRITE_RESP": Struct("result" / Flag),
        },
        default=Bytes(this.length),
    )
)


Command = Enum(
    Int8ub,
    RMU_DATA_MSG=0x62,
    MODBUS_DATA_MSG=0x68,
    MODBUS_READ_REQ=0x69,
    MODBUS_READ_RESP=0x6A,
    MODBUS_WRITE_REQ=0x6B,
    MODBUS_WRITE_RESP=0x6C,
)


# fmt: off
Response = Struct(
    "start_byte" / Const(0x5C, Int8ub),
    "empty_byte" / Const(0x00, Int8ub),
    "fields" / RawCopy(
        Struct(
            "address" / Enum(
                Int8ub,
                # 0x13 = 19, ?
                SMS40=0x16,
                RMU40=0x19,
                MODBUS40=0x20,
            ),
            "cmd" / Command,
            "length" / Int8ub,
            "data" / FixedSized(this.length, Data),
        )
    ),
    "checksum" / Checksum(Int8ub, xor8, this.fields.data),
).compile()


ReadRequest = Struct(
    "fields" / RawCopy(
        Struct(
            "start_byte" / Const(0xC0, Int8ub),
            "cmd" / Const(Command.MODBUS_READ_REQ, Command),
            "length" / Const(0x02, Int8ub),
            "coil_address" / Int16ul,
        )
    ),
    "checksum" / Checksum(Int8ub, xor8, this.fields.data),
).compile()


WriteRequest = Struct(
    "fields" / RawCopy(
        Struct(
            "start_byte" / Const(0xC0, Int8ub),
            "cmd" / Const(Command.MODBUS_WRITE_REQ, Command),
            "length" / Const(0x06, Int8ub),
            "coil_address" / Int16ul,
            "value" / Bytes(4),
        )
    ),
    "checksum" / Checksum(Int8ub, xor8, this.fields.data),
).compile()
# fmt: on



from typing import Dict, Optional, Union

from construct import (ConstructError, Int8sl, Int8ul, Int16sl, Int16ul, Int32sl,
                       Int32ul, Padded,)

from nibe.exceptions import DecodeException, EncodeException

parser_map = {
    "u8": Int8ul,
    "u16": Int16ul,
    "u32": Int32ul,
    "s8": Int8sl,
    "s16": Int16sl,
    "s32": Int32sl,
}


def is_coil_boolean(coil):
    if coil.factor != 1:
        return False

    if coil.min == 0 and coil.max == 1:
        return True

    if coil.mappings and all(k in ["0", "1"] for k in coil.mappings):
        return True

    return False


class Coil:
    mappings: Optional[Dict[str, str]]
    reverse_mappings: Optional[Dict[str, str]]

    def __init__(
        self,
        address: int,
        name: str,
        title: str,
        size: str,
        factor: int = 1,
        info: str = None,
        unit: str = None,
        mappings: dict = None,
        write: bool = False,
        **kwargs,
    ):
        assert isinstance(address, int), "Address must be defined"
        assert name, "Name must be defined"
        assert title, "Title must be defined"
        assert factor, "Factor must be defined"
        assert not (
            mappings is not None and factor != 1
        ), "When mapping is used factor needs to be 1"

        self.parser = parser_map.get(size)
        assert self.parser is not None

        self.address = address
        self.name = name
        self.title = title
        self.factor = factor

        self.set_mappings(mappings)

        self.info = info
        self.unit = unit
        self.is_writable = write

        self.other = kwargs

        self.raw_min = self.other.get("min")
        self.raw_max = self.other.get("max")

        self.min = self.raw_min / factor if self.raw_min is not None else None
        self.max = self.raw_max / factor if self.raw_max is not None else None

        self.is_boolean = is_coil_boolean(self)
        if self.is_boolean and not mappings:
            self.set_mappings({"0": "OFF", "1": "ON"})

        self._value = None

    def set_mappings(self, mappings):
        if mappings:
            self.mappings = dict((k, v.upper()) for k, v in mappings.items())
            self.reverse_mappings = dict((v.upper(), k) for k, v in mappings.items())
        else:
            self.mappings = None
            self.reverse_mappings = None

    @property
    def value(self) -> Union[int, float, str]:
        return self._value

    @value.setter
    def value(self, value: Union[int, float, str]):
        if self.mappings:
            value = value.upper()
            assert (
                value in self.reverse_mappings
            ), f"Provided value '{value}' is not in {self.reverse_mappings.keys()} for {self.name}"

            self._value = value
            return

        assert isinstance(
            value, (int, float)
        ), f"Provided value '{value}' is invalid type (int and float are supported) for {self.name}"

        self.check_value_bounds(value)

        self._value = value

    @property
    def raw_value(self) -> bytes:
        return self._encode(self.value)

    @raw_value.setter
    def raw_value(self, raw_value: bytes):
        self.value = self._decode(raw_value)

    def _decode(self, raw: bytes) -> Union[int, float, str]:
        value = self.parser.parse(raw)
        try:
            self._check_raw_value_bounds(value)
        except AssertionError as e:
            raise DecodeException(e)
        if self.factor != 1:
            value /= self.factor
        if self.mappings is None:
            return value

        mapped_value = self.mappings.get(str(value))
        if mapped_value is None:
            raise DecodeException(
                f"Mapping not found for {self.name} coil for value: {value}"
            )

        return mapped_value

    def _encode(self, val: Union[int, float, str]) -> bytes:
        try:
            if self.reverse_mappings is not None:
                mapped_value = self.reverse_mappings.get(str(val))
                if mapped_value is None:
                    raise EncodeException(
                        f"Mapping not found for {self.name} coil for value: {val}"
                    )

                return self._pad(mapped_value)

            if self.factor != 1:
                val *= self.factor

            self._check_raw_value_bounds(val)

            return self._pad(val)
        except AssertionError as e:
            raise EncodeException(e)
        except ConstructError as e:
            raise EncodeException(
                f"Failed to encode {self.name} coil for value: {val}, exception: {e}"
            )

    def _pad(self, value) -> bytes:
        return Padded(4, self.parser).build(int(value))

    def check_value_bounds(self, value):
        if self.min is not None:
            assert (
                value >= self.min
            ), f"{self.name} coil value is smaller than min({self.min}) allowed"

        if self.max is not None:
            assert (
                value <= self.max
            ), f"{self.name} coil value is larger than max({self.max}) allowed"

    def _check_raw_value_bounds(self, value):
        if self.raw_min is not None:
            assert (
                value >= self.raw_min
            ), f"{self.name} coil raw value is smaller than min({self.raw_min}) allowed"

        if self.raw_max is not None:
            assert (
                value <= self.raw_max
            ), f"{self.name} coil raw value is larger than max({self.raw_max}) allowed"

    def __repr__(self):
        return f"Coil {self.address}, name: {self.name}, title: {self.title}, value: {self.value}"



class NibeException(Exception):
    pass


class CoilNotFoundException(NibeException):
    pass


class DecodeException(NibeException):
    pass


class EncodeException(NibeException):
    pass


class CoilWriteException(NibeException):
    pass


class CoilWriteTimeoutException(CoilWriteException):
    pass


class CoilReadException(NibeException):
    pass


class CoilReadTimeoutException(CoilReadException):
    pass



import asyncio
import logging

from async_modbus import modbus_for_url

from nibe.coil import Coil
from nibe.connection import Connection
from nibe.exceptions import CoilReadException, CoilWriteException, DecodeException
from nibe.heatpump import HeatPump

logger = logging.getLogger("nibe").getChild(__name__)


class Modbus(Connection):
    DEFAULT_TIMEOUT = 5

    def __init__(self, heatpump: HeatPump, url, slave_id, conn_options=None):
        self._slave_id = slave_id
        self._heatpump = heatpump
        self._client = modbus_for_url(url, conn_options)

    async def read_coil(self, coil: Coil, timeout: float = DEFAULT_TIMEOUT) -> Coil:
        logger.debug(f"Sending read request")
        try:
            result = await asyncio.wait_for(
                self._client.read_input_registers(
                    slave_id=self._slave_id, starting_address=coil.address, quantity=1
                ),
                timeout,
            )
            coil.raw_value = result[0]
            logger.info(f"{coil.name}: {coil.value}")
            self._heatpump.notify_coil_update(coil)
        except asyncio.TimeoutError:
            raise CoilReadException(
                f"Timeout waiting for read response for {coil.name}"
            )

        return coil

    async def write_coil(self, coil: Coil, timeout: float = DEFAULT_TIMEOUT) -> Coil:
        assert coil.is_writable, f"{coil.name} is not writable"
        assert coil.value is not None

        logger.debug(f"Sending write request")
        try:
            result = await asyncio.wait_for(
                self._client.write_register(
                    slave_id=self._slave_id, address=coil.address, value=coil.raw_value
                ),
                timeout,
            )

            if not result:
                raise CoilWriteException(f"Heatpump denied writing {coil.name}")
            else:
                logger.info(f"Write succeeded for {coil.name}")
        except asyncio.TimeoutError:
            raise CoilWriteException(
                f"Timeout waiting for write feedback for {coil.name}"
            )

        return coil



import json
import logging
from collections import defaultdict
from enum import Enum
from importlib.resources import files
from typing import Any, Callable, Dict, Union

from nibe.coil import Coil
from nibe.exceptions import CoilNotFoundException

logger = logging.getLogger("nibe").getChild(__name__)


class Model(Enum):
    F1155 = "f1155_f1255"
    F1255 = "f1155_f1255"

    F1145 = "f1145_f1245"
    F1245 = "f1145_f1245"

    F1345 = "f1345"
    F1355 = "f1355"

    F730 = "f730"
    F750 = "f750"

    F370 = "f370_f470"
    F470 = "f370_f470"

    SMO20 = "smo20"
    SMO40 = "smo40"

    VVM225 = "vvm225_vvm320_vvm325"
    VVM320 = "vvm225_vvm320_vvm325"
    VVM325 = "vvm225_vvm320_vvm325"

    VVM310 = "vvm310_vvm500"
    VVM500 = "vvm310_vvm500"

    def get_coil_data(self):
        return json.loads(files("nibe.data").joinpath(f"{self.value}.json").read_text())

    @classmethod
    def keys(cls):
        return cls.__members__.keys()


class HeatPump:
    COIL_UPDATE_EVENT = "coil_update"

    _listeners: defaultdict[Any, list[Callable[..., None]]]
    _address_to_coil: Dict[str, Coil]
    _name_to_coil: Dict[str, Coil]

    def __init__(self, model: Model):
        assert isinstance(model, Model)
        self.model = model

        self._listeners = defaultdict(list)

    def _load_coils(self):
        data = self.model.get_coil_data()

        self._address_to_coil = {k: Coil(address=int(k), **v) for k, v in data.items()}
        self._name_to_coil = {c.name: c for _, c in self._address_to_coil.items()}

    def initialize(self):
        self._load_coils()

    def get_coil_by_address(self, address: Union[int, str]) -> Coil:
        try:
            return self._address_to_coil[str(address)]
        except KeyError:
            raise CoilNotFoundException(f"Coil with address {address} not found")

    def get_coil_by_name(self, name: str) -> Coil:
        try:
            return self._name_to_coil[str(name)]
        except KeyError:
            raise CoilNotFoundException(f"Coil with name '{name}' not found")

    def notify_coil_update(self, coil: Coil):
        for listener in self._listeners[self.COIL_UPDATE_EVENT]:
            try:
                listener(coil)
            except Exception as e:
                logger.exception(e)

    def subscribe(self, event_name: str, callback: Callable[..., None]):
        self._listeners[event_name].append(callback)



from abc import ABC, abstractmethod

from nibe.coil import Coil


class Connection(ABC):
    DEFAULT_TIMEOUT = 5.0

    async def start(self):
        pass

    async def stop(self):
        pass

    @abstractmethod
    async def read_coil(self, coil: Coil, timeout: float = DEFAULT_TIMEOUT) -> Coil:
        pass

    @abstractmethod
    async def write_coil(self, coil: Coil, timeout: float = DEFAULT_TIMEOUT) -> Coil:
        pass
